#include <ESP8266WiFi.h>
#include <WiFiUDP.h>
#include <EEPROM.h>

extern "C" {
#include "user_interface.h"
}

#define DEVICENUMBER 123 //wartość zapisywana do pamięci EEPROM pod adresem 0, jeżeli przy następnym uruchomieniu ta wartość nie zostanie tam znaleziona,
//do pamięci zostaną wgrane wartości domyślne

typedef union
  {
    uint32_t baudrate;
    struct
    {
      uint8_t byte1;
      uint8_t byte2;
      uint8_t byte3;
      uint8_t byte4;
    };
  } baudrate_unia; //unia zamieniająca 32bit wartość baudrate na 4 8bitowe zmienne, gotowe do zapisu w pamięci EEPROM

WiFiUDP Udp;

//port na którym odbędzie się komunikacja
unsigned int localPort = 8080;    

byte  packetBuffer[512]; 
int noBytes = 0;
String received_command = "";

int at_package = 0;
char at_data[50];
int at_cnt = 0;

char *Tx_IP;
uint16_t Tx_PORT = 8080;

//Adresy do zapisu zmiennych w EEPROM
int SW_Eeprom_Addr = 0;
int Baud_Eeprom_Addr = 1;
int Baud_set_Eeprom_Addr = 5;

//Flagi sprawdzające ustawienie odpowiednich parametrów
int Tx_IP_set = 0;
//int SSID_set = 0;
//int PASS_set = 0;
int Baud_set = 0;

//z powodu że funkcja indexOf przenosi wskaźnik w stringu na element szukany albo na koniec, trzeba operować na kopii danych
int find_text(String szukany, String zrodlo)
{
  return zrodlo.indexOf(szukany);
}

//ustawianie pamięci EEPROM
void EEprom_Configure()
{
  if (DEVICENUMBER != EEPROM.read(SW_Eeprom_Addr))
  {
    EEPROM.write(SW_Eeprom_Addr, DEVICENUMBER);
    EEPROM.write(Baud_set_Eeprom_Addr, Baud_set);
    EEPROM.commit();
  }
  else
  {
    Baud_set = EEPROM.read(Baud_set_Eeprom_Addr);
  }
}

//operacje na pamięci EEPROM
void EEprom_Baud_Save(uint32_t Baud)
{
  baudrate_unia baud;
  baud.baudrate = Baud;
  EEPROM.write(Baud_Eeprom_Addr, baud.byte1);
  EEPROM.write(Baud_Eeprom_Addr + 1, baud.byte2);
  EEPROM.write(Baud_Eeprom_Addr + 2, baud.byte3);
  EEPROM.write(Baud_Eeprom_Addr + 3, baud.byte4);
  EEPROM.write(Baud_set_Eeprom_Addr, Baud_set);
  EEPROM.commit();

}

uint32_t EEprom_Baud_Read()
{
  baudrate_unia baud;
  baud.byte1 = EEPROM.read(Baud_Eeprom_Addr);
  baud.byte2 = EEPROM.read(Baud_Eeprom_Addr + 1);
  baud.byte3 = EEPROM.read(Baud_Eeprom_Addr + 2);
  baud.byte4 = EEPROM.read(Baud_Eeprom_Addr + 3);

  return baud.baudrate;
}

//funkcja sprawdzająca czy w buforze danych znajduje się polecenie +AT
void AT_Parse(char* data)
{
  String AT_data(data);

  if (find_text("AT_Set_Recv_IP", AT_data) != -1)
  {
    String Reciver_IP;
    int poz1 = AT_data.indexOf("\"");
    int poz2 = AT_data.indexOf("\"", poz1 + 1);
    Reciver_IP = AT_data.substring(poz1 + 1, poz2);
    Tx_IP = new char[Reciver_IP.length() + 1];
    Reciver_IP.toCharArray(Tx_IP, Reciver_IP.length() + 1);
    Tx_IP[strlen(Tx_IP)] = '\0';
    Tx_IP_set = 1;
    Serial.println();
    Serial.println("OK");
    return;
  }
  else if (find_text("AT_Connect", AT_data) != -1)
  {
    String ssid_temp;
    String pass_temp;
    int i = 0;

    int poz1 = AT_data.indexOf("\"");
    int poz2 = AT_data.indexOf("\"", poz1 + 1);
    int poz3 = AT_data.indexOf("\"", poz2 + 1);
    int poz4 = AT_data.indexOf("\"", poz3 + 1);
    ssid_temp = AT_data.substring(poz1 + 1, poz2);
    pass_temp = AT_data.substring(poz3 + 1, poz4);
    char ssid[ssid_temp.length()];
    char pass[pass_temp.length()];
    ssid_temp.toCharArray(ssid, ssid_temp.length() + 1);
    pass_temp.toCharArray(pass, pass_temp.length() + 1);

    WiFi.begin(ssid, pass);


    Serial.println(ssid);
    Serial.println(pass);

    while (WiFi.status() != WL_CONNECTED && i++ < 20) delay(500);
    if (i == 21) {
      Serial.print("ERROR");
      return;
    }
    Udp.begin(localPort);
    Serial.println();
    Serial.println("OK");
    return;
  }
  else if (find_text("AT_Recv_IP", AT_data) != -1)
  {
    Serial.print("Recv IP:");
    Serial.println(Tx_IP);
    Serial.println();
    Serial.println("OK");
    return;
  }
  else if (find_text("AT_IP", AT_data) != -1)
  {
    Serial.print("Local IP:");
    Serial.println(WiFi.localIP());
    Serial.println();
    Serial.println("OK");
    return;

  }
  if (find_text("AT_Port", AT_data) != -1)
  {
    Serial.print("Port:");
    Serial.println(localPort);
    Serial.println();
    Serial.println("OK");
    return;
  }
  else if (find_text("AT_Set_Baud", AT_data) != -1)
  {
    String Baud;
    int poz1 = AT_data.indexOf("\"");
    int poz2 = AT_data.indexOf("\"", poz1 + 1);
    Baud = AT_data.substring(poz1 + 1, poz2);
    Baud_set = 1;
    EEprom_Baud_Save(Baud.toInt());
    delay(500);
    system_restart();

    Serial.println();
    Serial.println("OK");
    return;
  }
  else if (find_text("AT_Baud", AT_data) != -1)
  {
    Serial.print("Baud:");
    Serial.println(EEprom_Baud_Read());
    Serial.println();
    Serial.println("OK");
    return;
  }
  else if (find_text("AT_RST", AT_data) != -1)
  {
    system_restart();
  }
  else if (find_text("AT", AT_data) != -1)
  {
    Serial.println();
    Serial.println("OK");
    return;
  }
  else
  {
    Serial.println();
    Serial.println("ERROR");
  }

}

void setup() {

  uint8_t i = 0;

  EEPROM.begin(512);
  EEprom_Configure();
  //Jeżeli już był ustawiony baudrate, odczytaj go z EEPROM, jeżeli nie użyj domyślnej prędkośći
  if (Baud_set)
  {
    Serial.begin(EEprom_Baud_Read());
  }
  else
  {
    Serial.begin(115200);
  }
}

void loop() {

  noBytes = Udp.parsePacket();
  if ( noBytes ) {
    received_command = "";
    Udp.read(packetBuffer, noBytes);
    for (int i = 1; i <= noBytes; i++)
    {
      received_command = received_command + char(packetBuffer[i - 1]);
    } // end for

    Serial.println(received_command);
    Serial.println();
  }
  //Jeżeli jakieś dane czekają na odbiór na porcie szeregowym odczytaj je i sprawdź czy zaczynają się od +AT, jeżeli nie to wyślij je jako pakiet UDP, 
  //w innym przypadku zbierz dane aż do znaku '\n' i przekaż do funkcji parse
  if (Serial.available()) {
    size_t len = Serial.available();

    char data;

    if (!at_package)
    {
      data = Serial.read();
    }
    uint8_t sbuf[len];
    if (data == '+' && at_package == 0)
    {
      at_package = 1;
      at_cnt = 0;
      at_data[at_cnt++] = data;
      while (Serial.available())
      {
        at_data[at_cnt++] = Serial.read();
        //        Serial.print(at_data[at_cnt - 1]);
        if (at_data[at_cnt - 1] == '\n')
        {
          at_package = 0;
          AT_Parse(at_data);
          break;
        }
      }
    }
    else if (at_package)
    {
      while (len)
      {
        at_data[at_cnt++] = Serial.read();
        //        Serial.print(at_data[at_cnt - 1]);
        if (at_data[at_cnt - 1] == '\n')
        {
          at_package = 0;
          AT_Parse(at_data);
          break;
        }
        len--;
      }
    }
    //IP odbiornika musi być ustawione, inaczej nic nie zostanie wysłane
    else if (Tx_IP_set)
    {
      sbuf[0] = data;
      Serial.readBytes(&sbuf[1], len - 1);
      //push UART data to all connected telnet clients
      Udp.beginPacket(Tx_IP, Tx_PORT);
      Udp.write(sbuf, len);
      Udp.endPacket();
    }
    else
    {
      Serial.println("ERROR");
    }
  }

}